/*
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (c) 2011, 2012 digitalSTROM.org, Zurich, Switzerland
 *  Author: Sebastian Scholz <digitalstrom@sunshine-studios.org>
 */

Ext.define('DSS.addon.Metering', {

    extend : 'DSS.addon.Framework',

    uses : ['DSS.addon.Metering.MainPanel', 'Ext.Component'],

    config : {
        appName : _('Metering'),
        appId : 'metering',
        appVersion : '0.2.8',
        appIcon : 'images/dss/default_icon.png',
        appLang : 'de_DE'
    },

    getHelp : function() {
        return Ext.create('Ext.Component', {
            loader : {
                url : 'locale/' + this.appLang + '/LC_MESSAGES/help.html',
                autoLoad : true,
                disableCaching : false
            }
        });
    },
    getContent : function() {
        return Ext.create("DSS.addon.Metering.MainPanel", this.config);
    }
});

Ext.onReady(function() {
    var metering = Ext.create('DSS.addon.Metering');
    metering.initPage();
});
/*######################################
 static configs
 ######################################*/

DSS.addon.Metering.time = {
    1 : _('secondly'),
    60 : _('minutely'),
    900 : _('15 minutely'),
    86400 : _('daily'),
    604800 : _('weekly'),
    2592000 : _('monthly')
};

DSS.addon.Metering.steps = {
    1 : [Ext.Date.MINUTE, 1],
    60 : [Ext.Date.HOUR, 2],
    900 : [Ext.Date.HOUR, 6],
    86400 : [Ext.Date.DAY, 7],
    604800 : [Ext.Date.MONTH, 3],
    2592000 : [Ext.Date.MONTH, 4]
};

DSS.addon.Metering.minorTicks = {
    1 : 0,
    60 : 3,
    900 : 4,
    86400 : 6,
    604800 : 4,
    2592000 : 8
};

DSS.addon.Metering.dateFormat = {
    1 : 'G:i',
    60 : 'G:i',
    900 : 'd. M - G:i',
    86400 : 'd. M',
    604800 : 'd. M Y',
    2592000 : 'd. M Y'
};

DSS.addon.Metering.sceneNames = {
    0 : _("Off"),
    5 : _("Scene 1"),
    17 : _("Scene 2"),
    18 : _("Scene 3"),
    19 : _("Scene 4"),
    1 : _("Area 1 off"),
    6 : _("Area 1 on"),
    2 : _("Area 2 off"),
    7 : _("Area 2 on"),
    3 : _("Area 3 off"),
    8 : _("Area 3 on"),
    4 : _("Area 4 off"),
    9 : _("Area 4 on"),

    32 : _("Off (Scene 10)"),
    33 : _("Scene 11"),
    20 : _("Scene 12"),
    21 : _("Scene 13"),
    22 : _("Scene 14"),

    34 : _("Off (Scene 20)"),
    35 : _("Scene 21"),
    23 : _("Scene 22"),
    24 : _("Scene 23"),
    25 : _("Scene 24"),

    36 : _("Off (Scene 30)"),
    37 : _("Scene 31"),
    26 : _("Scene 32"),
    27 : _("Scene 33"),
    28 : _("Scene 34"),

    38 : _("Off (Scene 40)"),
    39 : _("Scene 41"),
    29 : _("Scene 42"),
    30 : _("Scene 43"),
    31 : _("Scene 44"),

    67 : _("Standby"),
    64 : _("Auto Standby"),
    68 : _("Deep off"),
    69 : _("Sleeping"),
    70 : _("Wake up"),
    72 : _("Leave home"),
    71 : _("Coming home"),
    73 : _("Door bell"),
    65 : _("Panic")
};

DSS.addon.Metering.refreshInterval = 2000;

DSS.addon.Metering.YAxisMax = 20;

/*######################################
 MainPanel
 ######################################*/
Ext.define('DSS.addon.Metering.MainPanel', {
    extend : 'Ext.panel.Panel',

    alias : 'widget.mainpanel',

    initComponent : function() {

        this.dSMPanel = Ext.create('DSS.addon.Metering.dSMPanel');
        this.centerPanel = Ext.create('DSS.addon.Metering.CenterPanel');

        Ext.apply(this, {
            //title : _('Metering Charts'),
            split : true,
            layout : {
                type : 'border'
            },
            items : [this.dSMPanel, this.centerPanel]
        });

        DSS.addon.Metering.MainPanel.superclass.initComponent.apply(this, arguments);
    }
});

/*######################################
 dSMPanel
 ######################################*/
Ext.define('DSS.addon.Metering.dSMPanel', {
    extend : 'Ext.panel.Panel',

    alias : 'widget.dsmpanel',

    initComponent : function() {

        this.dSMView = Ext.create('DSS.addon.Metering.dSMView');

        Ext.apply(this, {
            region : 'west',
            border : true,
            split : true,
            minWidth : 50,
            width : 200,
            layout : {
                type : 'fit'
            },
            items : [this.dSMView]
        });

        DSS.addon.Metering.dSMPanel.superclass.initComponent.apply(this, arguments);
    }
});

/*######################################
 dSMView
 ######################################*/
Ext.define('DSS.addon.Metering.dSMView', {
    extend : 'Ext.view.View',

    initComponent : function() {
        var me = this;
        this.circuitStore = Ext.create('DSS.addon.Metering.stores.CircuitViewStore');
        //TODO: remove the following comment as soon as metering data for the
        // whole installation is available
        /*this.circuitStore.add({
         name : _("whole installation"),
         dsid : 0
         });*/

        this.circuitStore.load();

        var circuitTemplate = new Ext.XTemplate('<tpl for=".">', '<div class="zone-wrap" id="zone-{dsid}">', '<span>', '{text}', '</span>', '</div>', '</tpl>', '<div class="x-clear"></div>');

        Ext.apply(this, {
            title : _("dSM Overview"),
            store : this.circuitStore,
            tpl : circuitTemplate,
            autoScroll : true,
            overflow : 'auto',
            singleSelect : true,
            multiSelect : false,
            simpleSelect : false,
            trackOver : true,
            overItemClass : 'x-item-over',
            overClass : 'x-item-over',
            itemSelector : 'div.zone-wrap',
            emptyText : _("loading dSMs"),
            listeners : {
                selectionchange : function(dv, nodes) {
                    this.getSelectedDSID();
                    this.findParentByType('dsmpanel').findParentByType('mainpanel').centerPanel.repaint();
                }
            }

        });
        this.refresh();

        DSS.addon.Metering.dSMView.superclass.initComponent.apply(this, arguments);
    },
    getSelectedDSID : function() {
        if(this.getSelectionModel().getCount() === 0) {
            this.getSelectionModel().select(0);
        }

        var node = this.getSelectionModel().getSelection();
        if(node !== null) {
            var dsid = node[0].get('dsid');
            if(dsid === 0) {
                //whole apartment is selected -> select next dsm
                this.getSelectionModel().select(1);
                node = this.getSelectionModel().getSelection();
                if(node !== null) {
                    dsid = node[0].get('dsid');
                }
            }
            return dsid;
        } else {
            Ext.MessageBox.alert(_('Error'), _('please select a dSM!'));
        }

        return 0;
    },
    getSelectedName : function() {
        if(this.getSelectionModel().getCount() === 0) {
            this.getSelectionModel().select(0);
        }

        var node = this.getSelectionModel().getSelection();
        if(node !== null) {
            var dsid = node[0].get('text');
            if(dsid === 0) {
                //whole apartment is selected -> select next dsm
                this.getSelectionModel().select(1);
                node = this.getSelectionModel().getSelection();
                if(node !== null) {
                    dsid = node[0].get('text');
                }
            }
            return dsid;
        } else {
            Ext.MessageBox.alert(_('Error'), _('please select a dSM!'));
        }

        return 0;
    }
});

/*######################################
 CenterPanel
 ######################################*/

Ext.define('DSS.addon.Metering.CenterPanel', {
    extend : 'Ext.panel.Panel',

    initComponent : function() {
        var me = this;

        this.unit = '';

        this.resolutionStore = Ext.create('DSS.addon.Metering.stores.ResolutionStore');
        this.resolutionStore.load();

        if(this.resolutionStore.isLoading()) {
            this.resolutionStore.on('load', function() {
                me.initStores();
            });
        } else {
            this.initStores();
        }

        this.sceneNameStore = Ext.create('DSS.addon.Metering.stores.SceneNameStore');
        this.sceneNameStore.load();

        //this store is used to display the right events in the realtime charts
        this.apartmentStore = Ext.create('DSS.apartment.ApartmentStore');
        this.apartmentStore.load();

        this.typeStore = Ext.create('Ext.data.Store', {
            fields : ['type', 'name'],
            data : [{
                "type" : "consumption",
                "name" : _('Consumption')
            }, {
                "type" : "energy",
                "name" : _('Energy')
            }]
        });

        this.typeBox = Ext.create('Ext.form.field.ComboBox', {
            store : this.typeStore,
            labelAlign : 'top',
            editable : false,
            displayField : 'name',
            valueField : 'type',
            fieldLabel : _('unit')
        });
        this.typeBox.on('change', function() {
            me.repaint();
        });

        this.timeBox = Ext.create('Ext.form.field.ComboBox', {
            store : this.resolutionStore,
            labelAlign : 'top',
            editable : false,
            queryMode : 'remote',
            lastQuery : '',
            forceSelection : true,
            displayField : 'name',
            valueField : 'resolution',
            fieldLabel : _('resolution')
        });
        this.timeBox.on('change', function() {
            me.repaint();
        });

        this.meteringLineStore = Ext.create('DSS.addon.Metering.MeteringStore');
        this.meteringColumnStore = Ext.create('DSS.addon.Metering.MeteringStore');
        this.meteringLineStore.on('load', function() {
            me.lineChart.axes.get(0).maximum = Math.max(DSS.addon.Metering.YAxisMax, me.meteringLineStore.max('value') * 1.2);
            me.lineChart.refresh();
        });
        this.meteringColumnStore.on('load', function() {
            me.columnChart.axes.get(1).fromDate = me.meteringColumnStore.min('date');
            me.columnChart.axes.get(1).toDate = me.meteringColumnStore.max('date');
            me.columnChart.axes.get(0).maximum = Math.max(DSS.addon.Metering.YAxisMax, me.meteringColumnStore.max('value') * 1.2);
            me.columnChart.refresh();
        });

        this.meteringSeries = Ext.create('DSS.addon.Metering.LineSeries');

        this.lineChart = Ext.create('Ext.chart.Chart', {
            store : this.meteringLineStore,
            autoScroll : true,
            zoomX : 0,
            axes : [{
                title : _('consumption') + ' (W)',
                type : 'Numeric',
                position : 'left',
                fields : ['value'],
                decimals : 0,
                maximum : Math.max(DSS.addon.Metering.YAxisMax, me.meteringLineStore.max('value') * 1.2),
                minimum : 0
            }, {
                type : 'Time',
                position : 'bottom',
                fields : 'date',
                dateFormat : 'd. M - G:i',
                dashSize : 6,

                setDateFormat : function(dateFormat) {
                    var me = this, label, f;
                    label = me.label || {};
                    label.renderer = function(v) {
                        return Ext.Date.format(new Date(v), dateFormat);
                    };
                }
            }],
            series : [this.meteringSeries],
            theme : 'MeteringTheme',

            originalFrom : null,
            originalToo : null,

            setZoom : function(zoomConfig) {
                var meToo = this, axes = meToo.axes, bbox = meToo.chartBBox, xScale = 1 / bbox.width, yScale = 1 / bbox.height, zoomer = {
                    x : zoomConfig.x * xScale,
                    y : zoomConfig.y * yScale,
                    width : zoomConfig.width * xScale,
                    height : zoomConfig.height * yScale
                };

                var axis = axes.get(1);
                if(!meToo.originalFrom) {
                    meToo.originalFrom = meToo.store.min('date');
                }
                if(!meToo.originalTo) {
                    meToo.originalTo = meToo.store.max('date');
                }
                var ends = axis.calcEnds();
                var from = (ends.to - ends.from) * zoomer.x + ends.from, to = (ends.to - ends.from) * zoomer.width + from;
                axis.minimum = from;
                axis.maximum = to;

                meToo.redraw(false);
            },
            restoreZoom : function(redraw) {
                var meToo = this;
                var axis = meToo.axes.get(1);
                if(meToo.originalFrom && meToo.originalTo) {
                    var ends = Ext.draw.Draw.snapEnds(meToo.orinalFrom, meToo.originalTo, meToo.majorTickSteps !== false ? (meToo.majorTickSteps + 1) : meToo.steps);

                    axis.minimum = ends.from;
                    axis.maximum = ends.to;

                    meToo.originalFrom = null;
                    meToo.orignalTo = null;
                }

                if(redraw) {
                    meToo.redraw(false);
                }
            },
            // override bindStore method so that the chart is not refreshed
            // automatically
            bindStore : function(store, initial) {
                var me = this;
                if(!initial && me.store) {
                    if(store !== me.store && me.store.autoDestroy) {
                        me.store.destroyStore();
                    } else {
                        me.store.un('datachanged', me.refresh, me);
                        me.store.un('add', me.delayRefresh, me);
                        me.store.un('remove', me.delayRefresh, me);
                        me.store.un('update', me.delayRefresh, me);
                        me.store.un('clear', me.refresh, me);
                    }
                }
                if(store) {
                    store = Ext.data.StoreManager.lookup(store);
                }
                me.store = store;
                if(store && !initial) {
                    me.refresh();
                }
            }
        });

        this.lineChart.on('refresh', function() {
            me.lineChart.series.get(0).updateMarkers();
        });

        this.columnChart = Ext.create('Ext.chart.Chart', {
            store : this.meteringColumnStore,
            axes : [{
                title : _('energy') + ' (Wh)',
                type : 'Numeric',
                position : 'left',
                fields : ['value'],
                decimals : 0,
                minimum : 0,
                maximum : Math.max(DSS.addon.Metering.YAxisMax, me.meteringColumnStore.max('value'))
            }, {
                type : 'Time',
                position : 'bottom',
                fields : 'date',
                constrain : true,
                dateFormat : 'd. M - G:i',
                dashSize : 6,

                setDateFormat : function(dateFormat) {
                    var me = this, label, f;
                    label = me.label || {};
                    label.renderer = function(v) {
                        return Ext.Date.format(new Date(v), dateFormat);
                    };
                },
                //override the function since it is not working if the label is
                // corrupt
                drawHorizontalLabels : function() {
                    var me = this, labelConf = me.label, floor = Math.floor, max = Math.max, axes = me.chart.axes, position = me.position, inflections = me.inflections, ln = inflections.length, labels = me.labels, labelGroup = me.labelGroup, maxHeight = 0, ratio, gutterY = me.chart.maxGutter[1], ubbox, bbox, point, prevX, prevLabel, projectedWidth = 0, textLabel, attr, textRight, text, label, last, x, y, i, firstLabel;
                    last = ln - 1;
                    //get a reference to the first text label dimensions
                    point = inflections[0];
                    firstLabel = me.getOrCreateLabel(0, me.label.renderer(labels[0]));
                    ratio = Math.floor(Math.abs(Math.sin(labelConf.rotate && (labelConf.rotate.degrees * Math.PI / 180) || 0)));

                    for( i = 0; i < ln; i++) {
                        point = inflections[i];
                        // here is my change
                        // skip label if label can't be rendered
                        if(labels[i] === Infinity) {
                            continue;
                        }
                        text = me.label.renderer(labels[i]);
                        textLabel = me.getOrCreateLabel(i, text);
                        bbox = textLabel._bbox;
                        maxHeight = max(maxHeight, bbox.height + me.dashSize + me.label.padding);
                        x = floor(point[0] - ( ratio ? bbox.height : bbox.width) / 2);
                        if(me.chart.maxGutter[0] == 0) {
                            if(i == 0 && axes.findIndex('position', 'left') == -1) {
                                x = point[0];
                            } else if(i == last && axes.findIndex('position', 'right') == -1) {
                                x = point[0] - bbox.width;
                            }
                        }
                        if(position == 'top') {
                            y = point[1] - (me.dashSize * 2) - me.label.padding - (bbox.height / 2);
                        } else {
                            y = point[1] + (me.dashSize * 2) + me.label.padding + (bbox.height / 2);
                        }

                        textLabel.setAttributes({
                            hidden : false,
                            x : x,
                            y : y
                        }, true);

                        // Skip label if there isn't available minimum space
                        if(i != 0 && (me.intersect(textLabel, prevLabel) || me.intersect(textLabel, firstLabel))) {
                            textLabel.hide(true);
                            continue;
                        }
                        prevLabel = textLabel;
                    }

                    return maxHeight;
                }
            }],
            series : [{
                type : 'column',
                xField : 'date',
                yField : 'value',
                tips : {
                    trackMouse : true,
                    width : 140,
                    height : 35,
                    renderer : function(storeItem, item) {
                        this.setTitle(storeItem.get('value').toFixed(0) + 'Wh ' + _('at') + ' ' + Ext.Date.format(new Date(storeItem.get('date')), 'd.m.Y - G:i:s'));
                    }
                }
            }],
            theme : 'MeteringTheme',

            originalFrom : null,
            originalToo : null,

            setZoom : function(zoomConfig) {
                var meToo = this, axes = meToo.axes, bbox = meToo.chartBBox, xScale = 1 / bbox.width, yScale = 1 / bbox.height, zoomer = {
                    x : zoomConfig.x * xScale,
                    y : zoomConfig.y * yScale,
                    width : zoomConfig.width * xScale,
                    height : zoomConfig.height * yScale
                };

                var axis = axes.get(1);
                if(!meToo.originalFrom) {
                    meToo.originalFrom = meToo.store.min('date');
                }
                if(!meToo.originalTo) {
                    meToo.originalTo = meToo.store.max('date');
                }
                var ends = axis.calcEnds();
                var from = (ends.to - ends.from) * zoomer.x + ends.from, to = (ends.to - ends.from) * zoomer.width + from;
                axis.minimum = from;
                axis.maximum = to;

                meToo.redraw(false);
            },
            restoreZoom : function(redraw) {
                var meToo = this;
                var axis = meToo.axes.get(1);
                if(meToo.originalFrom && meToo.originalTo) {
                    var ends = Ext.draw.Draw.snapEnds(meToo.orinalFrom, meToo.originalTo, meToo.majorTickSteps !== false ? (meToo.majorTickSteps + 1) : meToo.steps);

                    axis.minimum = ends.from;
                    axis.maximum = ends.to;

                    meToo.originalFrom = null;
                    meToo.orignalTo = null;
                }

                if(redraw) {
                    meToo.redraw(false);
                }
            },
            // override bindStore method so that the chart is not refreshed
            // automatically
            bindStore : function(store, initial) {
                var me = this;
                if(!initial && me.store) {
                    if(store !== me.store && me.store.autoDestroy) {
                        me.store.destroyStore();
                    } else {
                        me.store.un('datachanged', me.refresh, me);
                        me.store.un('add', me.delayRefresh, me);
                        me.store.un('remove', me.delayRefresh, me);
                        me.store.un('update', me.delayRefresh, me);
                        me.store.un('clear', me.refresh, me);
                    }
                }
                if(store) {
                    store = Ext.data.StoreManager.lookup(store);
                }
                me.store = store;
                if(store && !initial) {
                    me.refresh();
                }
            }
        });

        this.zoomSlider = Ext.create('Ext.slider.Slider', {
            width : 150,
            value : 100,
            increment : 10,
            minValue : 100,
            maxValue : 800,
            useTips : false,

            reset : function() {
                me.zoomSlider.setValue(100);
            }
        });

        this.zoomSlider.on('changecomplete', function() {
            var box = me.lineChart.chartBBox;
            me.lineChart.zoomX = Math.min(me.lineChart.zoomX, box.width - box.width / (me.zoomSlider.getValue() / 100));
            me.setChartZoom();
        });

        this.incrementZoomButton = Ext.create('Ext.button.Button', {
            cls : 'x-button-transparent',
            icon : 'images/dss/add.gif',
            handler : function() {
                me.zoomSlider.setValue(me.zoomSlider.getValue() + 50, true);
                me.setChartZoom();
            }
        });

        this.decrementZoomButton = Ext.create('Ext.button.Button', {
            cls : 'x-button-transparent',
            icon : 'images/dss/delete.png',
            handler : function() {
                me.zoomSlider.setValue(me.zoomSlider.getValue() - 50, true);
                me.setChartZoom();
            }
        });

        this.zoomFieldContainer = Ext.create('Ext.form.FieldContainer', {
            fieldLabel : _("Zoom"),
            labelAlign : 'top',
            layout : {
                type : 'table',
                columns : 3
            },
            items : [me.decrementZoomButton, me.zoomSlider, me.incrementZoomButton]
        });

        this.panSlider = Ext.create('Ext.slider.Slider', {
            width : '100',
            value : 0,
            increment : 5,
            minValue : 0,
            maxValue : 100,
            cls : 'x-slider-background',
            useTips : false,

            reset : function() {
                me.zoomSlider.setValue(0);
            }
        });

        this.panSlider.on('change', function() {
            me.setChartZoom();
        })

        this.exportCSVButton = Ext.create('Ext.button.Button', {
            text : _("export as CSV"),
            iconCls : 'icon-export',
            handler : function() {
                var dSMName = this.findParentByType('mainpanel').dSMPanel.dSMView.getSelectedName();
                var date = Ext.Date.format(new Date(), 'd-m-Y');

                // a temporary form has to be used so that the download
                var tempForm = document.createElement("form");
                tempForm.action = '/cgi-bin/pushfile.hs';
                tempForm.method = "POST";
                tempForm.style.display = "none";
                var params = {
                    Filename : dSMName + '_' + date + '.csv',
                    Contenttype : 'text/csv',
                    Data : me.exportCSV()
                };
                for(var par in params) {
                    var opt = document.createElement("textarea");
                    opt.name = par;
                    opt.value = params[par];
                    tempForm.appendChild(opt);
                }
                document.body.appendChild(tempForm);
                tempForm.submit();
                document.body.removeChild(tempForm);
            }
        });

        this.thread = null;
        this.repaintLock = true;

        Ext.apply(this, {
            layout : 'fit',
            items : this.lineChart,
            tbar : [this.typeBox, this.timeBox, {
                xtype : 'tbseparator'
            }, this.zoomFieldContainer, '->', this.exportCSVButton],
            bbar : [this.panSlider],
            region : 'center',
            border : true,
            split : true,
            minWidth : 600
        });

        me.on('resize', function() {
            me.panSlider.setWidth(me.getWidth() - 5);
        });

        DSS.addon.Metering.CenterPanel.superclass.initComponent.apply(this, arguments);
    },
    initStores : function() {
        for( i = 0; i < this.resolutionStore.getCount(); i++) {
            var record = this.resolutionStore.getAt(i);
            record.setName(DSS.addon.Metering.time[record.get('resolution')]);
        }
        this.typeBox.setValue('consumption');
        this.unit = 'W';

        this.resolutionStore.clearFilter();
        this.resolutionStore.filter({
            filterFn : function(item) {
                var resolution = item.get('resolution');
                if(resolution === 1 || resolution === 60 || resolution === 2592000) {
                    return true;
                }
                return false;
            }
        });
    },
    repaint : function() {
        if(this.repaintLock) {
            this.repaintLock = false;
            var me = this;
            var dsid = this.findParentByType('mainpanel').dSMPanel.dSMView.getSelectedDSID();
            var url;

            // stop running thread (not stopping it may lead to incorrect values
            // added to the store)
            this.stopThread();

            this.zoomSlider.reset();
            this.panSlider.reset();

            if(this.typeBox.getValue() === 'consumption') {
                //check if something is selected, otherwise select second entry
                // (first entry would enable realtime data)
                if(this.timeBox.getValue() === undefined) {
                    this.timeBox.setValue(this.resolutionStore.getAt(1).get('resolution'));
                }

                this.unit = 'W';

                //show only the specified values in the timeBox
                this.resolutionStore.clearFilter();
                this.resolutionStore.filter({
                    filterFn : function(item) {
                        var resolution = item.get('resolution');
                        if(resolution === 1 || resolution === 60 || resolution === 2592000) {
                            return true;
                        }
                        return false;
                    }
                });

                //only allow selected values that are in the store
                var value = this.timeBox.getValue();
                if((value !== 1 & value !== 60 & value !== 2592000)) {
                    this.timeBox.setValue(this.resolutionStore.getAt(1));
                }

                this.lineChart.axes.get(1).step = DSS.addon.Metering.steps[this.timeBox.getValue()];
                this.lineChart.axes.get(1).minorTickSteps = DSS.addon.Metering.minorTicks[this.timeBox.getValue()];
                this.lineChart.axes.get(1).setDateFormat(DSS.addon.Metering.dateFormat[this.timeBox.getValue()]);
                url = '/json/metering/getValues?dsid=' + dsid + '&type=consumption&resolution=' + this.timeBox.getValue() + '&valueCount=600';

                if(this.items.get(0) != this.lineChart) {
                    //show line chart, set column chart invisible due to
                    // performance reasons
                    this.columnChart.setVisible(false);
                    this.lineChart.setVisible(true);
                    this.remove(this.columnChart, false);
                    this.add(this.lineChart);
                }

                //elements might be disabled, because the column chart was
                // displayed
                me.zoomSlider.enable();
                me.decrementZoomButton.enable();
                me.incrementZoomButton.enable();
                me.panSlider.enable();

                this.lineChart.restoreZoom(false);

                this.meteringLineStore.removeAll();
                this.meteringLineStore.getProxy().url = url;
                this.meteringLineStore.load();

            } else {
                //check if something is selected, otherwise select first entry
                if(this.timeBox.getValue() === undefined) {
                    this.timeBox.setValue(this.resolutionStore.getAt(0).get('resolution'));
                }

                this.unit = 'Wh';

                //show only the specified values in the timeBox
                this.resolutionStore.clearFilter();
                this.resolutionStore.filter({
                    filterFn : function(item) {
                        var resolution = item.get('resolution');
                        if(resolution === 60 || resolution === 900 || resolution === 86400 || resolution === 604800 || resolution === 2592000) {
                            return true;
                        }
                        return false;
                    }
                });

                //only allow selected values that are in the store
                var value = this.timeBox.getValue();
                if((value !== 60 & value !== 900 & value !== 86400 & value !== 604800 & value !== 2592000)) {
                    this.timeBox.setValue(this.resolutionStore.getAt(0));
                }

                this.columnChart.axes.get(1).step = DSS.addon.Metering.steps[this.timeBox.getValue()];
                this.columnChart.axes.get(1).minorTickSteps = DSS.addon.Metering.minorTicks[this.timeBox.getValue()];
                this.columnChart.axes.get(1).setDateFormat(DSS.addon.Metering.dateFormat[this.timeBox.getValue()]);
                url = '/json/metering/getValues?dsid=' + dsid + '&type=energyDelta&resolution=' + this.timeBox.getValue() + '&valueCount=150';

                if(this.items.get(0) != this.columnChart) {
                    // show column chart, set line chart invisible due to
                    // performance reasons
                    this.lineChart.setVisible(false);
                    this.columnChart.setVisible(true);
                    this.remove(this.lineChart, false);
                    this.add(this.columnChart);
                }

                //no zooming for the column chart!
                me.zoomSlider.disable();
                me.decrementZoomButton.disable();
                me.incrementZoomButton.disable();
                me.panSlider.disable();

                this.columnChart.restoreZoom(false);

                this.meteringColumnStore.removeAll();
                this.meteringColumnStore.getProxy().url = url;
                this.meteringColumnStore.load();
            }

            if(this.typeBox.getValue() === 'consumption' && this.timeBox.getValue() === 1) {
                this.stopThread();
                this.startThread();
            } else {
                this.stopThread();
            }
            this.repaintLock = true;
        }
    },
    getUnit : function() {
        return this.unit;
    },
    setChartZoom : function() {
        var me = this;

        var box = me.lineChart.chartBBox;

        //set the x-border of the zoom area, be aware of the current zoom level
        me.lineChart.zoomX = Math.min((box.width - box.width / (me.zoomSlider.getValue() / 100)) * me.panSlider.getValue() / 100, box.width - box.width / (me.zoomSlider.getValue() / 100));

        if(me.lineChart.isVisible()) {
            var lineBox = me.lineChart.chartBBox;
            if(lineBox) {
                var sel = {
                    x : me.lineChart.zoomX,
                    y : 0,
                    width : lineBox.width / (me.zoomSlider.getValue() / 100),
                    height : lineBox.height
                };
                me.lineChart.restoreZoom();
                me.lineChart.setZoom(sel);
            }
        }

        // no zooming for the column chart at the moment
        /*if(me.columnChart.isVisible()) {
         var columBox = me.columnChart.chartBBox;
         if(columBox) {
         var sel = {
         x : 0,
         y : 0,
         width : columBox.width / (me.zoomSlider.getValue() / 100),
         height : columBox.height
         };
         me.columnChart.restoreZoom();
         me.columnChart.setZoom(sel);
         }
         }*/
    },
    exportCSV : function() {
        var csvString = '"' + _("Time") + '";' + '"' + _("Value") + ' (' + _("in") + ' ' + this.unit + ')"' + '\n', i = 0, r = null;
        if(this.typeBox.getValue() === 'consumption') {
            for( i = 0; i < this.meteringLineStore.getCount(); i++) {
                r = this.meteringLineStore.getAt(i);
                csvString += '"' + new Date(r.get('date')) + '";' + r.get('value') + '\n';
            }
        } else {
            for( i = 0; i < this.meteringColumnStore.getCount(); i++) {
                r = this.meteringColumnStore.getAt(i);
                csvString += '"' + new Date(r.get('date')) + '";' + r.get('value') + '\n';
            }
        }
        return csvString;
    },
    // starts the polling of the new metering data every x seconds via a
    // setInterval method and an Ajax request
    // also jumps in the data series are detected and jump markers are added to
    // the right position
    startThread : function() {
        var me = this;

        //no zooming if realtime data is displayed
        me.zoomSlider.disable();
        me.decrementZoomButton.disable();
        me.incrementZoomButton.disable();
        me.panSlider.disable();

        //detect jumps in the data of the store
        if(this.meteringLineStore.isLoading()) {
            this.meteringLineStore.on('load', function() {
                me.lineChart.axes.get(0).maximum = Math.max(DSS.addon.Metering.YAxisMax, me.meteringLineStore.max('value') * 1.2);
                var count = me.meteringLineStore.getCount();
                for(var i = 5; i < count; i++) {
                    var a = me.meteringLineStore.getAt(i - 5);
                    var b = me.meteringLineStore.getAt(i - 4);
                    var x = me.meteringLineStore.getAt(i - 3);
                    var c = me.meteringLineStore.getAt(i - 2);
                    var d = me.meteringLineStore.getAt(i - 1);

                    me.detectJumps(a, b, x, c, d);
                }
            }, me, {
                single : true
            });
        } else {
            me.lineChart.axes.get(0).maximum = Math.max(DSS.addon.Metering.YAxisMax, me.meteringLineStore.max('value') * 1.2);
            var count = me.meteringLineStore.getCount();
            for(var i = 5; i < count; i++) {
                var a = me.meteringLineStore.getAt(i - 5);
                var b = me.meteringLineStore.getAt(i - 4);
                var x = me.meteringLineStore.getAt(i - 3);
                var c = me.meteringLineStore.getAt(i - 2);
                var d = me.meteringLineStore.getAt(i - 1);

                me.detectJumps(a, b, x, c, d);
            }
        }

        this.thread = setInterval(function() {
            Ext.Ajax.request({
                url : '/json/metering/getLatest?type=consumption&from=.meters(' + me.findParentByType('mainpanel').dSMPanel.dSMView.getSelectedDSID() + ')',
                // set the timeout to the refresh interval since we will do
                // another request every 2 seconds
                timeout : DSS.addon.Metering.refreshInterval,
                success : function(response) {
                    var jsonData = Ext.JSON.decode(response.responseText);
                    if(jsonData.ok) {
                        var array = [];
                        //convert dateString to a format every browser
                        //should understand
                        var dateString = jsonData.result.values[0].date;
                        var date = Ext.Date.parse(dateString, 'Y-m-d g:i:s');
                        var record = [date, jsonData.result.values[0].value];
                        array.push(record);

                        me.meteringLineStore.removeAt(0);
                        me.meteringLineStore.loadData(array, true);

                        // adjust Y-range so that no buble is cut off
                        me.lineChart.axes.get(0).maximum = Math.max(me.lineChart.axes.get(0).maximum, jsonData.result.values[0].value * 1.2);

                        var count = me.meteringLineStore.getCount();
                        var a = me.meteringLineStore.getAt(count - 7);
                        var b = me.meteringLineStore.getAt(count - 6);
                        var x = me.meteringLineStore.getAt(count - 5);
                        var c = me.meteringLineStore.getAt(count - 4);
                        var d = me.meteringLineStore.getAt(count - 3);

                        me.detectJumps(a, b, x, c, d);

                        me.lineChart.refresh();
                    }
                }
            });
        }, DSS.addon.Metering.refreshInterval);
    },
    /**   detect jumps
     * the parameters of this function are storeItems
     * jump detection is based on the last 5 values
     *
     *              c----d
     *             /
     *            /
     *           x
     *          /
     *         /
     *   a----b
     *
     * a jump is detected for c, if
     *      |a-b| < 5
     *      |a-c| > 10
     *      |c-d| < 5
     *      |c-x| > 5
     *
     */
    detectJumps : function(itemA, itemB, itemX, itemC, itemD) {
        var me = this;

        var a = itemA.get('value');
        var b = itemB.get('value');
        var x = itemX.get('value');
        var c = itemC.get('value');
        var d = itemD.get('value');

        if(Math.abs(a - b) < 5 && Math.abs(c - d) < 5 && Math.abs(a - c) > 10 && Math.abs(c - x) > 5) {
            var diff = c - a;
            var diffString;
            if(diff > 0) {
                diffString = '+' + diff + ' ' + me.unit;
            } else {
                diffString = diff + ' ' + me.unit;
            }

            // load the last called scenes from the property tree
            Ext.Ajax.request({
                url : '/json/property/getString?path=/scripts/metering/lastCalledScenes',
                // set the timeout to the refresh interval since we will do
                // another request every 2 seconds
                timeout : DSS.addon.Metering.refreshInterval,
                success : function(response) {
                    var jsonData = Ext.JSON.decode(response.responseText);
                    if(jsonData.ok) {
                        var jsonString = jsonData.result.value;
                        var lastCalledScenes = Ext.JSON.decode(jsonString);

                        //set an interval of 10 seconds to compare the timestamp
                        // of the last called scene with
                        var now = new Date(itemD.get('date'));
                        var before = new Date(now);
                        before.setTime(before.getTime() - 10000);
                        var after = new Date(now);
                        after.setTime(after.getTime() + 10000);

                        var sceneNameFound = false;
                        var sceneName;

                        var zoneID;
                        var groupID;
                        var sceneID;

                        for( i = lastCalledScenes.length - 1; i >= 0; i--) {
                            var time = new Date(lastCalledScenes[i].time);
                            if(before <= time && time <= after) {
                                zoneID = parseInt(lastCalledScenes[i].zoneID, 10);
                                groupID = parseInt(lastCalledScenes[i].groupID, 10);
                                sceneID = parseInt(lastCalledScenes[i].sceneID, 10);
                                sceneName = me.getSceneName(sceneID, groupID, zoneID);
                                sceneNameFound = true;
                                break;
                            }
                        }

                        if(sceneNameFound) {
                            var sceneString;
                            if(zoneID !== 0) {
                                var zoneName = me.sceneNameStore.getAt(0).zones().getById(zoneID).get('text');
                                sceneString = sceneName + ' ' + _("in Zone") + ' ' + zoneName;
                            } else {
                                sceneString = sceneName;
                            }
                            me.lineChart.series.get(0).addJumpMarker(itemD, diffString, sceneString);
                        } else {
                            me.lineChart.series.get(0).addJumpMarker(itemD, diffString);
                        }
                    }
                }
            });
        }
    },
    stopThread : function() {
        if(this.thread != null) {
            clearInterval(this.thread);
            this.lineChart.series.get(0).removeAllMarkers();
            this.lineChart.redraw(false);
        }
    },
    getSceneName : function(sceneID, groupID, zoneID) {
        var me = this;
        var sceneName;

        //if the zone of the last scene call is 0 we don't have to check if the
        // zone is handled by the dsm
        if(zoneID === 0) {
            sceneName = DSS.addon.Metering.sceneNames[sceneID];
        } else if(groupID === 0) {
            sceneName = DSS.addon.Metering.sceneNames[sceneID];
        } else {
            //check if the zone of the matching scene call is handled by the
            // selected dSM
            var selectedDSID = me.findParentByType('mainpanel').dSMPanel.dSMView.getSelectedDSID();

            for( j = 0; j < me.apartmentStore.getAt(0).zones().getById(zoneID).devices().getCount(); j++) {
                var meterDSID = me.apartmentStore.getAt(0).zones().getById(zoneID).devices().getAt(0).get('meterDSID');
                if(meterDSID === selectedDSID) {
                    if(me.sceneNameStore.getAt(0).zones().getById(zoneID).groups().getById(groupID).scenes().getById(sceneID)) {
                        sceneName = me.sceneNameStore.getAt(0).zones().getById(zoneID).groups().getById(groupID).scenes().getById(sceneID).get('name');
                    } else {
                        sceneName = DSS.addon.Metering.sceneNames[sceneID];
                    }
                    break;
                }
            }
        }

        return sceneName;
    }
});

/*######################################
 MeteringLineSeries
 ######################################

 Behaves like a normal LineSeries, but adds some functions to draw markers
 to specific storeItems.
 */

Ext.define('DSS.addon.Metering.LineSeries', {

    type : 'Line',
    xField : 'date',
    yField : 'value',
    axis : ['left', 'bottom'],
    smooth : false,
    showMarkers : false,
    highlight : false,
    style : {
        fill : '#00523b',
        stroke : '#00523b',
        'stroke-width' : 3
    },

    constructor : function(config) {
        this.callParent(arguments);
        var me = this;
        Ext.apply(me, config, {
            tips : {
                trackMouse : true,
                width : 140,
                height : 35,
                renderer : function(storeItem, item) {
                    this.setTitle(storeItem.get('value').toFixed(0) + 'W ' + _('at') + ' ' + Ext.Date.format(new Date(storeItem.get('date')), 'd.m.Y - G:i:s'));
                }
            }
        });
    },
    // array to store all markers
    jumpMarkers : [],

    // add a new marker to the given store item. The text is displayed in the
    // chart
    addJumpMarker : function(storeItem, diff, sceneName, color) {
        var me = this, i = 0;

        for( i = 0, ln = me.items.length; i < ln; i++) {
            if(storeItem.get('date') == me.items[i].value[0] && storeItem.get('value') == me.items[i].value[1]) {
                var diffSprite = Ext.create('Ext.draw.Sprite', {
                    type : 'text',
                    myType : 'diff',
                    text : diff,
                    x : me.items[i].point[0] - 195,
                    y : me.items[i].point[1] - 46,
                    zIndex : 7000 + me.jumpMarkers.length * 10 + 5,
                    storeItem : storeItem
                });

                var frameSprite = Ext.create('Ext.draw.Sprite', {
                    type : 'path',
                    myType : 'frame',
                    path : "M 0 10 L 0 40 Q 0 50 10 50 L 170 50 L 200 60 L 180 40 L 180 10 Q 180 0 170 0 L 10 0 Q 0 0 0 10 Z",
                    zIndex : 7000 + me.jumpMarkers.length * 10,
                    stroke : '#00523b',
                    'stroke-width' : 3,
                    fill : '#baefd1',
                    opacity : 1,
                    storeItem : storeItem
                });

                frameSprite.setAttributes({
                    translate : {
                        x : me.items[i].point[0] - 200,
                        y : me.items[i].point[1] - 60
                    }
                }, true);

                me.chart.surface.add(frameSprite);
                me.chart.surface.add(diffSprite);
                frameSprite.show(true);
                diffSprite.show(true);

                //check if a matching scene call was found
                if(sceneName) {
                    var sceneSprite = Ext.create('Ext.draw.Sprite', {
                        type : 'text',
                        myType : 'scene',
                        text : sceneName,
                        x : me.items[i].point[0] - 195,
                        y : me.items[i].point[1] - 28,
                        zIndex : 7000 + me.jumpMarkers.length * 10 + 5,
                        storeItem : storeItem
                    });

                    me.chart.surface.add(sceneSprite);
                    sceneSprite.show(true);
                    me.jumpMarkers.push(sceneSprite);
                }

                me.jumpMarkers.push(frameSprite);
                me.jumpMarkers.push(diffSprite);
            }
        }
    },
    getPositionForStoreItem : function(storeItem) {
        var me = this;

        for(var i = 0, ln = me.items.length; i < ln; i++) {
            if(storeItem.get('date') == me.items[i].value[0] && storeItem.get('value') == me.items[i].value[1]) {
                return me.items[i].point
            }
        }
    },
    // sets all markers to the new position of the item. Necessary if a new item
    // is added to the chart
    updateMarkers : function() {
        var me = this;

        for(var i = 0; i < me.jumpMarkers.length; i++) {
            var sprite = me.jumpMarkers[i];
            var point = me.getPositionForStoreItem(sprite.storeItem);
            if(point === undefined) {
                me.jumpMarkers.splice(i, 1);
                sprite.remove();
            } else {
                if(sprite.myType === 'diff') {
                    sprite.setAttributes({
                        x : point[0] - 195,
                        y : point[1] - 46
                    }, true);
                } else if(sprite.myType === 'scene') {
                    sprite.setAttributes({
                        x : point[0] - 195,
                        y : point[1] - 28
                    }, true);
                } else {
                    sprite.setAttributes({
                        translate : {
                            x : point[0] - 200,
                            y : point[1] - 60
                        }
                    }, true);
                }
            }
        }
    },
    removeAllMarkers : function() {
        var me = this;
        for(var i = 0; i < me.jumpMarkers.length; i++) {
            me.jumpMarkers[i].remove();
        }
    }
});

/*######################################
 MeteringModel
 ######################################*/

Ext.define('DSS.addon.Metering.MeteringModel', {
    extend : 'Ext.data.Model',
    fields : [{
        name : 'date',
        type : 'date',
        dateFormat : 'timestamp',
        mapping : 0
    }, {
        name : 'value',
        mapping : 1
    }]
});

/*######################################
 MeteringStore
 ######################################*/

Ext.define('DSS.addon.Metering.MeteringStore', {
    extend : 'Ext.data.Store',

    model : 'DSS.addon.Metering.MeteringModel',
    proxy : {
        type : 'ajax',
        actionMethods : {
            read : 'GET'
        },
        pageParam : undefined,
        limitParam : undefined,
        startParam : undefined,
        reader : {
            type : 'array',
            root : 'result.values',
            successProperty : 'ok'
        }
    },
    sorters : [{
        property : 'date',
        direction : 'ASC'
    }],
    // override remove method of this store so that no 'datachanged' or 'remove'
    // event is fired
    // -> chart is only redrawn when a new value pair is added to the store
    // -> better performance
    remove : function(records, /* private */isMove) {
        if(!Ext.isArray(records)) {
            records = [records];
        }
        /*
         * Pass the isMove parameter if we know we're going to be re-inserting
         * this record
         */
        isMove = isMove === true;
        var me = this, sync = false, i = 0, length = records.length, isPhantom, index, record;

        for(; i < length; i++) {
            record = records[i];
            index = me.data.indexOf(record);

            if(me.snapshot) {
                me.snapshot.remove(record);
            }
            if(index > -1) {
                isPhantom = record.phantom === true;
                if(!isMove && !isPhantom) {
                    // don't push phantom records onto removed
                    me.removed.push(record);
                }
                record.unjoin(me);
                me.data.remove(record);
                sync = sync || !isPhantom;
            }
        }

        if(!isMove && me.autoSync && sync) {
            me.sync();
        }
    }
});

/*######################################
 MeteringTheme
 ######################################*/

Ext.define('Ext.chart.theme.MeteringTheme', {
    extend : 'Ext.chart.theme.Base',

    constructor : function(config) {
        var baseColor = '#005138';
        //light green (like the borders)
        //var background = '#99e8bc';
        //darker green (like dSM Overview)
        //var fill = '#048c40';
        //very dark green (like the dS Logo)
        var fill = '#00523b';

        this.callParent([Ext.apply({
            axis : {
                fill : baseColor,
                stroke : baseColor
            },
            axisLabelLeft : {
                fill : baseColor
            },
            axisLabelBottom : {
                fill : baseColor
            },
            axisTitleLeft : {
                fill : baseColor
            },
            axisTitleBottom : {
                fill : baseColor
            },
            colors : [fill],
            markerThemes : [{
                fill : baseColor,
                type : 'circle'
            }]
        }, config)]);
    }
});

Ext.apply(Ext.Date, {
    monthNames : [
        _("January"),
        _("February"),
        _("March"),
        _("April"),
        _("May"),
        _("June"),
        _("July"),
        _("August"),
        _("September"),
        _("October"),
        _("November"),
        _("December")
    ]
});
